@@ -16,10 +16,10 @@
 
 #include <config-kcrash.h>
 
-#include <signal.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
+#include <csignal>
+#include <cstdio>
+#include <cstdlib>
+#include <cstring>
 
 #include <qplatformdefs.h>
 #ifndef Q_OS_WIN
@@ -38,6 +38,7 @@
 
 #include <algorithm>
 #include <array>
+#include <chrono>
 #include <memory>
 
 #include <QDebug>
@@ -45,6 +46,9 @@
 #include <QFile>
 #include <QGuiApplication>
 #include <QLibraryInfo>
+#include <QOffscreenSurface>
+#include <QOpenGLContext>
+#include <QOpenGLFunctions>
 #include <QStandardPaths>
 #include <QThread>
 
@@ -53,32 +57,27 @@
 #endif
 
 #include "coreconfig_p.h"
+#include "exception_p.h"
 #include "metadata_p.h"
 
-// Copy from klauncher_cmds
-typedef struct {
-    long cmd;
-    long arg_length;
-} kcrash_launcher_header;
+// WARNING: do not use qGlobalStatics in here, they get destroyed too early on
+// shutdown and may inhibit crash handling in late-exit scenarios (e.g. when
+// a function local static gets destroyed by __cxa_finalize)
+#undef Q_GLOBAL_STATIC
 
-#define LAUNCHER_OK 4
-#define LAUNCHER_EXEC_NEW 12
+using namespace std::chrono_literals;
 
-namespace KCrash
-{
-KCRASH_EXPORT bool loadedByKdeinit = false;
-void setApplicationFilePath(const QString &filePath);
-// Create socket path to transfer ptrace scope and open connection
-}
 #ifdef Q_OS_LINUX
 static QByteArray s_socketpath;
 #endif
 
 struct Args {
+    Args() = default;
     ~Args()
     {
         clear();
     }
+    Q_DISABLE_COPY_MOVE(Args)
 
     void clear()
     {
@@ -128,45 +127,48 @@ static int s_originalSignal = -1;
 static QByteArray s_metadataPath;
 
 static std::unique_ptr<char[]> s_kcrashErrorMessage;
-Q_GLOBAL_STATIC(KCrash::CoreConfig, s_coreConfig)
 
-static void kcrashInitialize()
+namespace
 {
-    // Static because in some cases this is called multiple times
-    // but if an application had any of the bad cases we always want
-    // to skip the check
-    static bool doAutoInitKCrash = true;
+const KCrash::CoreConfig s_coreConfig;
 
-    if (!doAutoInitKCrash) {
-        return;
+std::unique_ptr<char[]> s_glRenderer; // the GL_RENDERER
+
+QString glRenderer()
+{
+    QOpenGLContext context;
+    QOffscreenSurface surface;
+    surface.create();
+
+    if (!context.create()) {
+        return {};
     }
 
-    QCoreApplication *app = QCoreApplication::instance();
-    if (!app) {
-        doAutoInitKCrash = false;
-        return;
+    if (!context.makeCurrent(&surface)) {
+        return {};
     }
 
-    if (!QCoreApplication::startingUp()) {
-        // If the app has already started, this means we're not being run as part of
-        // qt_call_pre_routines, which most probably means that we're being run as part
-        // of KCrash being loaded as part of some plugin of the app, so don't
-        // do any magic
-        doAutoInitKCrash = false;
-        return;
-    }
-
-    if (!QCoreApplication::eventDispatcher()) {
-        // We are called with event dispatcher being null when KCrash is being loaded
-        // through plasma-integration instead of being linked to the app (i.e. QtCreator vs Okular)
-        // For apps that don't link directly to KCrash do not do the magic
-        doAutoInitKCrash = false;
-        return;
-    }
-
-    KCrash::initialize();
+    auto done = qScopeGuard([&context] {
+        context.doneCurrent();
+    });
+    return QString::fromUtf8(reinterpret_cast<const char *>(context.functions()->glGetString(GL_RENDERER)));
 }
-Q_COREAPP_STARTUP_FUNCTION(kcrashInitialize)
+
+QString bootId()
+{
+#ifdef Q_OS_LINUX
+    QFile file(QStringLiteral("/proc/sys/kernel/random/boot_id"));
+    if (!file.open(QFile::ReadOnly)) {
+        qCWarning(LOG_KCRASH) << "Failed to read /proc/sys/kernel/random/boot_id" << file.errorString();
+        return {};
+    }
+    return QString::fromUtf8(file.readAll().simplified().replace('-', QByteArrayView()));
+#else
+    return {};
+#endif
+}
+
+} // namespace
 
 static QStringList libexecPaths()
 {
@@ -197,7 +199,7 @@ static bool shouldWriteMetadataToDisk()
     // the daemon may already be gone but we'll still want to deal with the crash on next login!
     // Similar reasoning applies to not checking the presence of the launcher socket.
     const bool drkonqiCoredumpHelper = !QStandardPaths::findExecutable(QStringLiteral("drkonqi-coredump-processor"), libexecPaths()).isEmpty();
-    return s_coreConfig()->isCoredumpd() && drkonqiCoredumpHelper && !qEnvironmentVariableIsSet("KCRASH_NO_METADATA");
+    return s_coreConfig.isCoredumpd() && drkonqiCoredumpHelper && !qEnvironmentVariableIsSet("KCRASH_NO_METADATA");
 #else
     return false;
 #endif
@@ -208,22 +210,31 @@ void KCrash::initialize()
     if (s_launchDrKonqi == 0) { // disabled by the program itself
         return;
     }
+
+    bool enableDrKonqi = !qEnvironmentVariableIsSet("KDE_DEBUG");
+    if (qEnvironmentVariableIsSet("KCRASH_AUTO_RESTARTED") || qEnvironmentVariableIntValue("RUNNING_UNDER_RR") == 1
+        || qEnvironmentVariableIntValue("KCRASH_DUMP_ONLY") == 1) {
+        enableDrKonqi = false;
+    }
+
     const QStringList args = QCoreApplication::arguments();
-    if (!qEnvironmentVariableIsSet("KDE_DEBUG") //
-        && !qEnvironmentVariableIsSet("KCRASH_AUTO_RESTARTED") //
-        && !qEnvironmentVariableIntValue("RUNNING_UNDER_RR") //
-        && qEnvironmentVariableIntValue("KCRASH_DUMP_ONLY") == 0) {
-        // enable drkonqi
+    // Default to core dumping whenever a process is set. When not or when explicitly opting into just in time debugging
+    // we enable drkonqi. This causes the signal handler to directly fork drkonqi opening us to race conditions.
+    // NOTE: depending on the specific signal other threads are running while the signal handler runs and may trip over
+    //   the signal handler's closed FDs. That is primarily why we do not like JIT debugging.
+    if (enableDrKonqi && (!s_coreConfig.isProcess() || qEnvironmentVariableIntValue("KCRASH_JIT_DRKONQI") == 1)) {
         KCrash::setDrKonqiEnabled(true);
     } else {
-        // This loads qtlogging.ini very early which prevents unittests from doing QStandardPaths::setTestModeEnabled(true) in initTestCase()
-        // qCDebug(LOG_KCRASH) << "KCrash disabled through environment.";
+        // Don't qDebug here, it loads qtlogging.ini very early which prevents unittests from doing QStandardPaths::setTestModeEnabled(true) in initTestCase()
     }
 
     if (QCoreApplication::instance()) {
         const QString path = QCoreApplication::applicationFilePath();
         s_appFilePath.reset(qstrdup(qPrintable(path))); // This intentionally cannot be changed by the application!
         KCrash::setApplicationFilePath(path);
+        if (auto guiApp = qobject_cast<QGuiApplication *>(QCoreApplication::instance())) {
+            s_glRenderer.reset(qstrdup(glRenderer().toUtf8().constData()));
+        }
     } else {
         qWarning() << "This process needs a QCoreApplication instance in order to use KCrash";
     }
@@ -236,9 +247,14 @@ void KCrash::initialize()
     if (shouldWriteMetadataToDisk()) {
         // We do not actively clean up metadata via KCrash but some other service. This potentially means we litter
         // a lot -> put the metadata in a subdir.
+        // This data is consumed by DrKonqi in combination with coredumpd metadata.
         const QString metadataDir = QStandardPaths::writableLocation(QStandardPaths::GenericCacheLocation) + QStringLiteral("/kcrash-metadata");
         if (QDir().mkpath(metadataDir)) {
-            s_metadataPath = QFile::encodeName(metadataDir + QStringLiteral("/%1.ini").arg(QCoreApplication::applicationPid()));
+            const auto bootId = ::bootId();
+            const auto exe = QString::fromUtf8(s_appName.get());
+            const auto pid = QString::number(QCoreApplication::applicationPid());
+            s_metadataPath = QFile::encodeName(metadataDir + //
+                                               QStringLiteral("/%1.%2.%3.ini").arg(exe, bootId, pid));
         }
         if (!s_crashHandler) {
             // Always enable the default handler. We cannot create the metadata ahead of time since we do not know
@@ -251,8 +267,6 @@ void KCrash::initialize()
             setCrashHandler(defaultCrashHandler);
         }
     } // empty s_metadataPath disables writing
-
-    s_coreConfig(); // Initialize.
 }
 
 void KCrash::setEmergencySaveFunction(HandlerType saveFunction)
@@ -284,7 +298,7 @@ class KCrashDelaySetHandler : public QObject
 public:
     KCrashDelaySetHandler()
     {
-        startTimer(10000); // 10 s
+        startTimer(10s);
     }
 
 protected:
@@ -315,7 +329,7 @@ void KCrash::setFlags(KCrash::CrashFlags flags)
 
 void KCrash::setApplicationFilePath(const QString &filePath)
 {
-    const int pos = filePath.lastIndexOf(QLatin1Char('/'));
+    const auto pos = filePath.lastIndexOf(QLatin1Char('/'));
     const QString appName = filePath.mid(pos + 1);
     const QString appPath = filePath.left(pos); // could be empty, in theory
 
@@ -383,26 +397,33 @@ void KCrash::setCrashHandler(HandlerType handler)
     sigset_t mask;
     sigemptyset(&mask);
 
+    const auto signals = {
 #ifdef SIGSEGV
-    signal(SIGSEGV, handler);
-    sigaddset(&mask, SIGSEGV);
+        SIGSEGV,
 #endif
 #ifdef SIGBUS
-    signal(SIGBUS, handler);
-    sigaddset(&mask, SIGBUS);
+        SIGBUS,
 #endif
 #ifdef SIGFPE
-    signal(SIGFPE, handler);
-    sigaddset(&mask, SIGFPE);
+        SIGFPE,
 #endif
 #ifdef SIGILL
-    signal(SIGILL, handler);
-    sigaddset(&mask, SIGILL);
+        SIGILL,
 #endif
 #ifdef SIGABRT
-    signal(SIGABRT, handler);
-    sigaddset(&mask, SIGABRT);
+        SIGABRT,
 #endif
+    };
+
+    for (const auto &signal : signals) {
+        struct sigaction action {
+        };
+        action.sa_handler = handler;
+        action.sa_flags = SA_RESTART;
+        sigemptyset(&action.sa_mask);
+        sigaction(signal, &action, nullptr);
+        sigaddset(&mask, signal);
+    }
 
     sigprocmask(SIG_UNBLOCK, &mask, nullptr);
 #endif
@@ -480,6 +501,19 @@ void KCrash::defaultCrashHandler(int sig)
         }
 #endif
 
+        if (auto optionalExceptionMetadata = KCrash::exceptionMetadata(); optionalExceptionMetadata.has_value()) {
+            if (optionalExceptionMetadata->klass) {
+                data.add("--exceptionname", optionalExceptionMetadata->klass);
+            }
+            if (optionalExceptionMetadata->what) {
+                data.add("--exceptionwhat", optionalExceptionMetadata->what);
+            }
+        }
+
+        if (s_glRenderer) {
+            data.add("--glrenderer", s_glRenderer.get());
+        }
+
         const QByteArray platformName = QGuiApplication::platformName().toUtf8();
         if (!platformName.isEmpty()) {
             data.add("--platform", platformName.constData());
@@ -500,10 +534,6 @@ void KCrash::defaultCrashHandler(int sig)
 
         data.add("--appname", s_appName ? s_appName.get() : "<unknown>");
 
-        if (loadedByKdeinit) {
-            data.addBool("--kdeinit");
-        }
-
         // only add apppath if it's not NULL
         if (s_appPath && s_appPath[0]) {
             data.add("--apppath", s_appPath.get());
@@ -555,21 +585,15 @@ void KCrash::defaultCrashHandler(int sig)
         const int argc = data.argc;
         const char **argv = data.argv.data();
 
-#ifndef NDEBUG
-        fprintf(stderr, "KCrash: crashing... crashRecursionCounter = %d\n", crashRecursionCounter);
-        fprintf(stderr,
-                "KCrash: Application Name = %s path = %s pid = %lld\n",
-                s_appName ? s_appName.get() : "<unknown>",
-                s_appPath ? s_appPath.get() : "<unknown>",
-                QCoreApplication::applicationPid());
-        fprintf(stderr, "KCrash: Arguments: ");
-        for (int i = 0; i < s_autoRestartCommandLine.argc; ++i) {
-            fprintf(stderr, "%s ", s_autoRestartCommandLine.argv[i]);
-        }
-        fprintf(stderr, "\n");
-#else
-        fprintf(stderr, "KCrash: Application '%s' crashing...\n", s_appName ? s_appName.get() : "<unknown>");
+        fprintf(stderr, "KCrash: Application '%s' crashing... crashRecursionCounter = %d\n", s_appName ? s_appName.get() : "<unknown>", crashRecursionCounter);
+
+        if (s_launchDrKonqi != 1) {
+            setCrashHandler(nullptr);
+#if !defined(Q_OS_WIN)
+            raise(sig); // dump core, or whatever is the default action for this signal.
 #endif
+            return;
+        }
 
 #if !defined(Q_OS_WIN) && !defined(Q_OS_OSX)
         if (!(s_flags & KeepFDs)) {
@@ -591,13 +615,18 @@ void KCrash::defaultCrashHandler(int sig)
 #endif
 #endif
 
-        if (s_launchDrKonqi != 1) {
-            setCrashHandler(nullptr);
-#if !defined(Q_OS_WIN)
-            raise(sig); // dump core, or whatever is the default action for this signal.
-#endif
-            return;
+#ifndef NDEBUG
+        fprintf(stderr,
+                "KCrash: Application Name = %s path = %s pid = %lld\n",
+                s_appName ? s_appName.get() : "<unknown>",
+                s_appPath ? s_appPath.get() : "<unknown>",
+                QCoreApplication::applicationPid());
+        fprintf(stderr, "KCrash: Arguments: ");
+        for (int i = 0; i < s_autoRestartCommandLine.argc; ++i) {
+            fprintf(stderr, "%s ", s_autoRestartCommandLine.argv[i]);
         }
+        fprintf(stderr, "\n");
+#endif
 
         startProcess(argc, argv, true);
     }
@@ -606,7 +635,7 @@ void KCrash::defaultCrashHandler(int sig)
         fprintf(stderr, "Unable to start Dr. Konqi\n");
     }
 
-    if (s_coreConfig->isProcess()) {
+    if (s_coreConfig.isProcess()) {
         fprintf(stderr, "Re-raising signal for core dump handling.\n");
         KCrash::setCrashHandler(nullptr);
         raise(sig);
@@ -712,7 +741,7 @@ void KCrash::startProcess(int argc, const char *argv[], bool waitAndExit)
             // If the process was started directly, use waitpid(), as it's a child...
             while (waitpid(pid, nullptr, 0) != pid) { }
         }
-        if (!s_coreConfig->isProcess()) {
+        if (!s_coreConfig.isProcess()) {
             // Only exit if we don't forward to core dumps
             _exit(253);
         }
